/*
 *     Copyright CERN, Geneva 1989 - Copyright and any other
 *     appropriate legal protection of these computer programs
 *     and associated documentation reserved in all countries
 *     of the world.
 */

/*
 *  Copyright (c) 1984 Marvin Solomon; all rights reserved
 */
/*
 * Global definitions for user telnet
 *
 *
 *
 */
#include "globals.h"

/*
 * If either GRAPHICS or TTY are defined then we will need transparent I/O.
 */

#ifdef GRAPHICS

/*
 * flag for transparent graphics feature (BS)
 *      > 0 : waiting for transparent input
 *      < 0 : transparent output mode
 */

int     graphics = 0;

#endif /* GRAPHICS */

#ifdef TTY

/* flag for kermit (transparent) input mode */

int     kermit = 0;

#endif /* TTY */

#ifdef TRANSP_IO

/*
 * Overall flag non-zero if in transparent I/O mode
 *      > 0 : waiting for transparent input
 *      < 0 : transparent output mode
 */

int     trflag = 0;

#endif /* TRANSP_IO */

char    *ttyn;
int     ttypid;

char    *term_type, termcap[1024];
char    *keymap_type;

/* character buffer for a full file name */
char    full_fname[100] = "";

char    eschar[20] = "";                   /* (BS) */

/* pointers to buffer for holding a 3270 command from the host */
unsigned char *savebuf = 0;     /* base address of allocated buffer */
unsigned char *savep = 0;       /* pointer within allocated buffer  */

/* current size of savebuf */

int     saveb_size = 0;

/* file descriptors */

/* Raw output device file descriptor and stream FILE structure */

int     fd_output;
FILE    *fd_stream;

int     net = 0;        /* file descriptor for the network connection */

#ifdef FROMNET
int     fromnet =  0;   /* file descriptor for logging everying from the network */
#endif /* FROMNET */

#ifdef TONET
int     tonet = 0;      /* file descriptor for logging everying to the network */
#endif /* TONET */

#ifdef DEBUG
FILE    *outf = NULL;   /* debug output file                    */
#endif /* DEBUG */

#ifdef LDEBUG
FILE    *log = NULL;    /* for logging various interesting things */
#endif /* LDEBUG */

/* Flag for indicating that the program is ending (in end_program) */

int     ending_program = 0;

/* Flag when we want synchronous output */

int     synch_op = SYNCH_NORMAL;

/*
 * Flag for saying what is the mode of the user's terminal
 *  0:  We have still got his original terminal setup
 *  1:  Now he is set up in (asynchronous) 3270 setup
 * -1:  Now he is temporarily back in (synchronous) dialogue mode
 */

int     term_mode = 0;

/*
 * Flag which will be set when the IBM sets telnet binary mode
 * 0: in "glass tty" mode
 * 1: in 3270 emulation mode
 */

int     binflag = 0;

/* states of various options */
int     will_tt, will_eor, do_eor, will_binary, do_binary, do_sga, will_tm;
int     sent_tt; /* IAC SB TERMINAL_TYPE IS ... IAC SE has been sent */

/*
 * There follow the various options that a user can have which can be
 * set in the call or changed in the on-line help options menu.
 * The first, however, is special: it identifies the last ANNOUNCE file
 * which the user has seen.
 * Note that the OPT_COUNT symbol must match the total number of options.
 */

/* Flag for the last ANNOUNCE file seen by the user */

int     announce_id = -1;

/* Flag for if 7171 emulation is required */

int     emul7171 = -1;

/* Flag for if quiet mode wanted */

int     bells_mode = -1;

/* Flag for if we should show MDT fields, possibly in reverse video */

int     show_MDT_field = -1;

#ifdef OPTDIR

/* File name for the user's chosen options */

char    opt_file[100] = "default_options";

/* FILE structure pointer for the option file (NULL if non-existent) */

FILE    *opt_fstr = NULL;

#endif /* OPTDIR */

/*
 * The addresses of the various option variables (all non-negative integers)
 * plus the associated format statements to read or write them.
 * Note that the first one MUST be the positive identification value
 * for the last announcement file that the user has seen.
 */

int     *opt_pointer[OPT_COUNT] = {
					&announce_id,
					&emul7171,
					&bells_mode,
					&show_MDT_field
				  };

int     opt_default[OPT_COUNT] = {
					-1,
					1,
					1,
					0
				 };

#ifdef OPTDIR

char    *opt_format[OPT_COUNT] =  {
					"announce_id = %d\n",
					"emul7171 = %d\n",
					"bells_mode = %d\n",
					"show_MDT_field = %d\n"
				  };

#endif /* OPTDIR */

/* type of virtual 3270 terminal */

char    *TERMNAME;      /* IBM-3278-x */
int     rows;           /* 24, 32, or 43 */
int     linelen;        /* 80  or 132 */
int     screensize;

char    *progname;      /* to remember argv[0] (BS)  */
char    *remotename;    /* to remember required host name/number (JMG) */
char    *username=NULL; /* to remember (optional) IBM user name (JMG) */
char    *dialname=NULL; /* to remember (optional) dial host name */

#ifdef LOGFILE
int     name_field = -1;/* field for initial name to be logged */
#endif /* LOGFILE */

/*
 * Tables for input escape sequences.
 *
 * Escape_char[] and escape_ptr are parallel arrays simulating an
 * array of structs (but 25% smaller) containing a trie.  Each node in
 * the trie is represented by a contiguous segment of the arrays,
 * terminated by a slot with escape_char[i] = '\377'.
 * The entries of a node are sorted by the escape_char field.  Each node
 * corresponds to the set of all strings with a given initial prefix.
 * Consider slot i in the node corressponding to prefix 'w',
 * with escape_char[i]=c and escape_ptr[i]=j.  If j>0, j is the start
 * of the node corresponding to wc.  If j<0, -j is the action code associated
 * with the string wc.
 *
 * Escape_start[c] points to the start of the node corresponding to the
 * (one-character) prefix 'c'.  If no string starts with 'c', escape_start[c]
 * is null.
 *
 * Note that this representation precludes '\377' as a character in a
 * <string> (except as the first character).
 */

short           escape_start[256];
unsigned char   escape_char[MAXESCAPECHARS];
short           escape_ptr[MAXESCAPECHARS];
int             next_escape;

char    *find_def();

/* Various 3270 indicators */
int     clearsn=0;      /* Positive if clear screen requested by IBM */
			/* It is 1 for normal screen, 2 for alternate */
			/* It is negated when the screen has been cleared, */
			/* but before the new IBM screen (if any) drawn. */
bool    insert=FALSE;   /* TRUE if terminal in insert mode */
bool    unlock=FALSE;   /* TRUE if terminal input is allowed */

/* status indicators as to whether things are working or hung */

int     op_wait=0;      /* # characters of output waiting to go to the user */

#ifdef TTY
int     idle_count= -1; /* count how long user has done nothing */
#endif /* TTY */


/*
 * Data structure for representing fields.
 * The screen array simulates the contents of 3278 internal memory.
 * It is allocated by the alloc_buffer routine.
 * Each position contains in the least significant byte a field-definition
 * character or an ASCII ISO 8859-1 character.
 * The upper part can have flags in it: FIELDSTART is the main one, so that
 * (screen[p] & FIELDSTART) != 0 iff position p is the start of a field.
 * In this case the lower part is copied from the attribute byte following
 * the SF order.
 *
 * For normal characters there is a flag GE_CHAR to indicate that the
 * character came after a Graphics Escape byte. In such cases the translation
 * to and from ASCII must go according to a different table. This table
 * will convert an EBCDIC Graphic character into a 16-bit character in
 * which the least significant 7 bits is the ASCII code, whilst the bits
 * above select one of several possible character sets.
 *
 * In order to have compatibility with an 8-bit system we will consider
 * the 0x80 bit as both part of an 8-bit character and the selection of
 * the required 7-bit character set when in a 7-bit environment.
 * The 0x100 bit may, however, select a different character set from
 * the ASCII ISO 8859-1 one. Thus, the combination of bits 0x100 and 0x80
 * can imply any of the following:-
 *
 * 00   ASCII 7-bit     normally assumed in G0.
 * 01   ASCII 8-bit     normally assumed in G2 (ISO Latin-1 if on a VT300
 *                      or clone, DEC supplemental if a VT200 or clone).
 * 10   Graphics 7-bit  normally assumed in G1 as DEC Special Graphics
 *                      (even exists on the good old VT100!).
 * 11   Graphics 8-bit  normally assumed in G3 as DEC Technical Character
 *                      set, but only on a VT300 or clone.
 *
 * Unfortunately, however, we may get the IBM sending characters which
 * are undefined in the graphics set (to be displayed as '-') or ones we
 * cannot print. To cater for these cases without losing the one-to-one
 * mapping property we include a bit to specify extra sets in which
 * every 8-bit character comes out looking like the same character. Thus,
 * the translation of any such character is the same character byte but
 * in this mythical character set. The reverse mapping is therefore
 * unique and identical. We put this bit into the 0x200 position, making
 * four extra 8-bit mappings of bits 0x200, 0x100 and 0x80 as follows :-
 *
 * 10x  Undefined graphics character: print as '-'
 * 11x  Unprintable graphics character: print as some special character.
 *
 */

short   *screen;

/*
 * To know which regions of the screen array need updating.
 * s_point is an array parallel to screen.
 * s_pfirst is either -1 or an index into s_point.
 *   if an index then s_point[s_pfirst] is also an index into s_point
 *   indicating that the screen region from s_point to screen[s_point]
 *   inclusive is new (to be written to the user screen).
 * Since there may be more than one such region we continue with
 *   screen[screen[s.point]] as either -1 or a new pointer.
 */

short   *s_point, s_pfirst = -1;

/* String to change from one color to another */

char    *change_cp[MAXCOLORS][MAXCOLORS];

/* Length of corresponding change_cp string */

int     change_sl[MAXCOLORS][MAXCOLORS];

/* Boolean to say if cursor can be moved when in a particular color */

bool    move_ok[MAXCOLORS];

/* Boolean to say whether blanks look "normal" in this color. */
/* If so then we can use the "clear to end of line" sequence. */

bool    ceol_ok[MAXCOLORS];

/* Boolean to say if the current screen has only NORMAL-style characters */

bool    normal_chars;

/* Boolean to say if the current screen has only NORMAL-style blanks */

bool    normal_blanks;


/* screen[firstfield] is the first element with the FIELDSTART bit set.
 * If no fields are currently defined, firstfield == -1.
 */
int     firstfield;

/* The contents of the screen are copied into the curscr data
 * structure maintained by curses.
 * curscr->char_y and curscr->ccset_y are arrays of pointers,
 * each pointing to an array of characters representing one line.
 * The char_y array is the actual character on the user screen.
 * The ccset_y array is both the color and the character set of the
 * character on the user screen.
 * This ccset is divided into two 4-bit "nibbles".
 * The more significant nibble is the "color" as one of the options
 * NORMAL, NORM_BLINK, ... , INVISIBLE
 * On simple terminals this is merely whether the character is highlighted
 * or not: more sophisticated terminals might have more properties such
 * as colour, blinking, dim etc.
 * The less significant nibble is the special indicator for
 * characters which came from the IBM with graphics escape and/or
 * are not part of the standard 8-bit ASCII set.
 */

unsigned char   **char_image, **ccset_image;

/* The contents of char_image differ from screen in the following ways:
 *   char_image is an array of arrays, while screen is one-dimensional
 *   The following kinds of positions contain spaces in char_image
 *      field definition characters
 *      positions in invisible fields
 *      null characters
 * ccset_image is like char_image, except that it is for the "colour and
 * character set" of the character, not the character itself.
 */

/* Extension to allow for both a normal and an alternate screen */

int     LINES;          /* Current number of lines */
int     COLS;           /* Current number of columns */

bool    altflip;        /* TRUE if there is an alternate screen  */
int     altscreen;      /* 0: normal screen, 1: alternate screen */

int     init_COLS;      /* lines on initial user display     */
int     init_LINES;     /* columns on initial user display   */

int     normal_linelen; /* lines in normal 3270 screen       */
int     normal_rows;    /* rows in normal 3270 screen        */
int     normal_LINES;   /* lines on normal user display      */
int     normal_COLS;    /* columns on normal user display    */
int     normal_slinenum;/* status line on normal display     */

int     altern_linelen; /* lines in alternate 3270 screen    */
int     altern_rows;    /* rows in alternate 3270 screen     */
int     altern_LINES;   /* lines on alternate user display   */
int     altern_COLS;    /* columns on alternate user display */
int     altern_slinenum;/* status line on normal display     */

short   ly, lx;         /* current cursor position */

/* The following variables keep track of the current cursor position */
int     cursorpos;      /* 0..screensize-1 */
int     cursorsave;     /* 0..screensize-1 */
int     bufferpos;      /* current buffer position (set by SBA order) 0..screensize-1 */
int     fieldstart;     /* first position of field currently containing cursor */
int     currentcolor;   /* color associated with the current field (fieldstart) */

/* the following variables specify new parts of screen for update_curscr */
int     startpos = -1;  /* start position (may be field definition) */
int     endpos;         /* end position (exclusive) */

/* keep count of calls to update_curscr (will be reset by put/writecursor */
int     updateflag;

bool    sline = FALSE;  /* TRUE if writing 3270 status line */
int     slinenum;       /* status line number if on normal screen */

int     scroll_line= -1;/* >0 : line which would cause scroll */
			/* <0 : any line COULD cause a scroll */

/*
 * Variables concerned with 8-bit mode, extended ASCII and so on.
 * Terminals set up for 7-bit mode can nonetheless have the extended
 * characters by the judicious use of Shift In/Out (SI/SO).
 * For 8-bit data input we may need to transform 8-bit control codes
 * into two 7-bit codes unless we have seen that 8-bit controls is OK.
 */

/* Flag for if terminal can handle 8-bit controls */

bool    eight_bc = FALSE;

/* Flag for if terminal can handle 8-bit data */

bool    eight_bd = FALSE;

/* Mask for OR of input character (0x80 after SO received, reset by SI) */

short   SO_imask = 0;

/* Used to change incoming 8-bit control character to ESC + 7-bit */

bool    ESC_iflag = FALSE;

/* Used to change outgoing ESC + 7-bit to 8-bit control character */

bool    ESC_oflag;

/*
 * Used before and after an extended character (>0x7f).
 * The two bits just before the 7-bit character will index into
 * this table.
 */

char    *SO_string[4] = {
			  "\0"    ,
			  "\033N" ,
			  "\016"  ,
			  "\033O"
			};
char    *SI_string[4] = {
			  "\0"    ,
			  "\0"    ,
			  "\017"  ,
			  "\0"
			};

/*
 * Chararacter to put out when the IBM has sent a graphics escape
 * character which is either illegal or one that we cannot manage
 * to produce on our ASCII-style screen.
 */

char    NBG_illegal = '-';
char    NBG_notgot  = '?';

/*
 * Various things concerned with the use of termcap sequences.
 * We keep a set of standard sequences, possibly including timing chars,
 * so as to avoid using low-level routines like tputs all the time.
 * We also invent a few new entries: if they are not in the actual
 * termcap entry it will not matter.
 */

struct  curwin  curscr_window;
unsigned char   *curscrch_y[MAXLINES];
unsigned char   *curscrccs_y[MAXLINES];
struct  curwin  *curscr;

