/*
 *     Copyright CERN, Geneva 1989 - Copyright and any other
 *     appropriate legal protection of these computer programs
 *     and associated documentation reserved in all countries
 *     of the world.
 */

/*
 * routines to handle status line I/O
 */

#include "telnet.h"
#include "3270.h"
#include "globals.h"

/* a copy of the status line */
char    slimage_array[133];

/* a saved copy of the status line */
char    slisave_array[133];

/* a line of blanks (enough for at least a full line */
char    blanks_array[133];

char    *badcmess = "NO: ILLEGAL CHARACTER!";
int     badcfl=1;       /* bad input character          */

#define BADCMSZ 22              /* size of badcmess             */
#define BADCOFF 11              /* offset of badcmess           */

char    *badwmess = "NO: PROTECTED FIELD!";
int     badwfl=2;       /* bad write in status line     */

#define BADWMSZ 22              /* size of badwmess             */
#define BADWOFF 11              /* offset of badwmess           */

char    *badimess = "NO MORE INSERTING!";
int     badifl=4;       /* bad write in status line     */

#define BADIMSZ 22              /* size of badwmess             */
#define BADIOFF 11              /* offset of badwmess           */

int     slflags = 0;    /* Contains flags for warnings in status line */
int     slfsave = 0;    /* Saved value of slflags after slsave        */
int     slchars;        /* Chars to put into status line              */

/* Now all of the status line routines */

/*
 * Put a string onto the status line at an offset position.
 * The bells parameter indicates how many times to ring the bell to warn
 * the user: 1 means an information message, >1 is a warning.
 * For an information message the bell will NOT be rung if the user
 * has a status line.
 * Note, however, that no bells will be rung if the user has switched
 * to quiet mode (bells_mode = 0).
 */

#ifdef STANDARD_C

void    slputstr(char *slstr, int offset, int bells)

#else /* STANDARD_C */

void    slputstr(slstr, offset, bells)

char    *slstr;
int     offset;
int     bells;

#endif /* STANDARD_C */

{
	int     slstr_len;      /* length of string to be inserted */
	int     offreal;        /* the real offset. */
	int     i;
	int     slstyle;        /* style of stuff to be put in */
	int     curflsave;      /* save current screen flags */
	int     n_bells;        /* the real number of bell rings */

#ifdef DEBUG
	(void)fprintf(outf, "slputstr    : %d bells: at offset %d put string\n<%s>\n",
		  bells, offset, slstr);
#endif /* DEBUG */

	n_bells = bells_mode * bells;

	offreal = (offset < 0 ? 0 : offset);
	slstr_len = strlen(slstr);
	if (sline)
	{
		if (bells > 1)
		{
			slstyle = STATL_BLINK;
		}
		else
		{
			slstyle = STATL;
			n_bells = 0;
		}
		curflsave = curscr->_color;
		if (slinenum)
		{

#ifdef DEBUG
	(void)fprintf(outf, "slputstr    : move from ly = %d, lx = %d, color = %d\n",
		  ly, lx, curscr->_color);
#endif /* DEBUG */

			domvcur(ly, lx, slinenum - 1, offreal);
			if (offset >= 0)
			{
				f_puts(change_cp[curscr->_color][slstyle], FALSE);
				curscr->_color = slstyle;
			}
			f_puts(slstr, FALSE);
			domvcur(slinenum - 1, offreal+slstr_len, ly, lx);
			if (offset >= 0)
			{
				f_puts(change_cp[curscr->_color][curflsave], FALSE);
				curscr->_color = curflsave;
			}
		}
		else
		{
			/*
			 * For a VT300 the sequence to select the host-writeable status line
			 * needs no parameters. However, some other sequences such as
			 * those on aixterm in an rs6000 expect a count of how far down
			 * the line to go.
			 * For simplicity we always go to the start then rewrite the start
			 * of the status line up to where the new stuff will go. We then
			 * add in the new stuff.
			 * However, some selection sequences actually blank out the status
			 * line, meaning that we then have to rewrite the rest of it. This
			 * can be a bit annoying, especially on a LAT-connected ASCII terminal
			 * when all that we are doing is inserting then deleting the
			 * character which says the IBM is unable to accept input.
			 * For safety we assume that the status line WILL be blanked out
			 * when selected by the TS sequence UNLESS the symbol NO_SL_CLEAR
			 * is defined.
			 * On systems running ASCII terminals, including the TAGIBM type
			 * of application, it is highly recommended to ensure that no
			 * termcap sequence includes the blanking out and to include the
			 * -DNO_SL_CLEAR option in the Makefile.
			 */

			/* f_puttcs(tcsTS, FALSE); */
			tputs(tgoto(TS, 0, 0), 1, f_putcs);

			if (offset >= 0)
			{
				f_puts(change_cp[curscr->_color][NORMAL], FALSE);
				f_puts(change_cp[NORMAL][STATL], FALSE);
			}
			slimage_array[offreal] = '\0';
			f_puts(slimage_array, FALSE);
			if (offset >= 0)
				f_puts(change_cp[STATL][slstyle], FALSE);
			f_puts(slstr, FALSE);
			if (offset >= 0)
				f_puts(change_cp[slstyle][STATL], FALSE);

			/*
			 * The next line is only needed if the ts string in the
			 * termcap entry includes clearing the status line.
			 * It is much faster if it does not.
			 */

#ifndef NO_SL_CLEAR

			f_puts(slimage_array+offreal+slstr_len, FALSE);

#endif /* NO_SL_CLEAR */

			if (offset >= 0)
			{
				f_puts(change_cp[STATL][NORMAL], FALSE);
				f_puts(change_cp[NORMAL][curscr->_color], FALSE);
			}
			f_puttcs(tcsFS, FALSE);
		}
		f_flush();
	}

	(void)strncpy(slimage_array + offreal , slstr , (size_t) slstr_len);

#ifdef DEBUG
	(void)fprintf(outf, "slputstr    : status line is now %d chars\n<%s>\n",
		  strlen(slimage_array), slimage_array);
#endif /* DEBUG */

	for (i = n_bells ; i ; i--)
		DING
	if (bells > 1)
		clear_kbin();
}

/*
 * put a number of blanks onto the status line at an offset position
 */

#ifdef STANDARD_C

void    slerase(int count, int offset, int bells)

#else /* STANDARD_C */

void    slerase(count, offset, bells)

int     count;
int     offset;
int     bells;

#endif /* STANDARD_C */

{
	blanks_array[count] = '\0';
	slputstr(blanks_array, offset, bells);
	blanks_array[count] = ' ';
}

/*
 * save status line image (for later restore)
 */

#ifdef STANDARD_C

void    slsave(void)

#else /* STANDARD_C */

void    slsave()

#endif /* STANDARD_C */

{
	int i;

	for (i = 0 ; i <= 132 ; i++)
		slisave_array[i] = slimage_array[i];
	slfsave = slflags;

#ifdef DEBUG
	(void)fprintf(outf, "slsave      : status line save\n");
	(void)fprintf(outf, "slsave      : saved line is now %d chars\n<%s>\n",
		  strlen(slisave_array), slisave_array);
#endif /* DEBUG */

	slisave_array[slchars] = ' ';
	slclear();
	return;
}

/*
 * restore saved status line and redraw it
 */

#ifdef STANDARD_C

void    slrestore(void)

#else /* STANDARD_C */

void    slrestore()

#endif /* STANDARD_C */

{
	slchars = (VW ? linelen : linelen-1);
	slimage_array[slchars] = '\0';
	slisave_array[slchars] = '\0';
	slputstr(slisave_array, 0, 0);
	slflags = slfsave;

#ifdef DEBUG
	(void)fprintf(outf, "slrestore   : status line restore\n");
	(void)fprintf(outf, "slrestore   : restored line is now %d chars\n<%s>\n",
		  strlen(slimage_array), slimage_array);
#endif /* DEBUG */

	return;
}

/*
 * clear status line
 */

#ifdef STANDARD_C

void    slclear(void)

#else /* STANDARD_C */

void    slclear()

#endif /* STANDARD_C */

{
	slerase(slchars, -1, 0);
}

/*
 * refresh status line
 */

#ifdef STANDARD_C

void    slrefresh(void)

#else /* STANDARD_C */

void    slrefresh()

#endif /* STANDARD_C */

{
	int i;
	char    slicopy_array[133];

#ifdef DEBUG
      (void)fprintf(outf, "slrefresh   : slchars %d, linelen %d VW %d\n",
			      slchars,linelen,VW);
#endif /* DEBUG */

	slimage_array[slchars] = ' ';
	slchars = (VW ? linelen : linelen-1);
	slimage_array[slchars] = '\0';

	for (i = 0 ; i <= 132 ; i++)
		slicopy_array[i] = slimage_array[i];
	slputstr(slicopy_array, 0, 0);
}

/*
 * initialise status line
 */

#ifdef STANDARD_C

void    slinit(void)

#else /* STANDARD_C */

void    slinit()

#endif /* STANDARD_C */

{
	int i;

#ifdef DEBUG
	(void)fprintf(outf, "slinit      : status line initialise\n");
#endif /* DEBUG */

	for (i = 0 ; i <= 132 ; i++)
	{
		slimage_array[i] = ' ';
		blanks_array[i] = ' ';
	}
	slchars = (VW ? linelen : linelen-1);
	slimage_array[slchars] = '\0';
	if (sline)
	{
		if (!slinenum)
		{
			f_puttcs(tcsDS, FALSE);
		}
	}
	slputstr("ONLINE   ", 0 , 1);
	return;
}

/*
 * terminate status line usage
 */

#ifdef STANDARD_C

void    slterm(void)

#else /* STANDARD_C */

void    slterm()

#endif /* STANDARD_C */

{
	slclear();
	slinenum = 0;
	sline = FALSE;
}

/*
 *      clear any flags in status line
 */

#ifdef STANDARD_C

void    slfclear(void)

#else /* STANDARD_C */

void    slfclear()

#endif /* STANDARD_C */

{
	if (!slflags) return;

#ifdef DEBUG
	(void)fprintf(outf, "slfclear    : slflags entry %x\n",
		  slflags);
#endif /* DEBUG */

	if (slflags & badcfl)
	{
		slerase(BADCMSZ, BADCOFF, 0);
		slflags -= badcfl;
	}
	if (slflags & badwfl)
	{
		slerase(BADWMSZ, BADWOFF, 0);
		slflags -= badwfl;
	}
	if (slflags & badifl)
	{
		slerase(BADIMSZ, BADIOFF, 0);
		slflags -= badifl;
	}
}

/*
 *      Trying to input an illegal character.
 *      We may put something into the status line, which must later be
 *      cleared (by local action).
 */

#ifdef STANDARD_C

void    badchar(void)

#else /* STANDARD_C */

void    badchar()

#endif /* STANDARD_C */

{
		slflags |= badcfl;

#ifdef DEBUG
	(void)fprintf(outf, "badchar     : slflags is now %x\n",
		  slflags);
#endif /* DEBUG */

	slputstr(badcmess, BADCOFF, 2);
}

/*
 *      Trying to modify somewhere where it is not permitted.
 *      May be enter, insert, erase or delete a character when the cursor
 *      is in a protected field or an attribute location.
 *      We may put something into the status line, which must later be
 *      cleared (by local action).
 */

#ifdef STANDARD_C

void    badwrite(void)

#else /* STANDARD_C */

void    badwrite()

#endif /* STANDARD_C */

{
		slflags |= badwfl;

#ifdef DEBUG
	(void)fprintf(outf, "badwrite    : slflags is now %x\n",
		  slflags);
#endif /* DEBUG */

	slputstr(badwmess, BADWOFF, 2);
}

/*
 *      Trying to insert, but this would push characters into a protected
 *      field or attribute location.
 *      We may put something into the status line, which must later be
 *      cleared (by local action).
 */

#ifdef STANDARD_C

void    badinsert(void)

#else /* STANDARD_C */

void    badinsert()

#endif /* STANDARD_C */

{
		slflags |= badifl;

#ifdef DEBUG
	(void)fprintf(outf, "badinsert   : slflags is now %x\n",
		  slflags);
#endif /* DEBUG */

	slputstr(badimess, BADIOFF, 2);
}
